import ee from "~/event-emitter";
import reqTimeout from "~/request-timeout";
import lazyConfig from "~/lazysizes-config";
import Hammer from "hammerjs";

const eventNamespace = "tgam-photo-slider";

// component photo slider
const slideHasInfoClass = "c-photo-slider--has-info";
const slideHasNavClass = "c-photo-slider--has-nav";
const slideHasShareToolsClass = "c-photo-slider--has-share-tools";

const slideListClass = "c-photo-slider__list";
const slideItemClass = "c-photo-slider__item";

const slideNavLeftClass = "c-photo-slider__nav--left";
const slideNavRightClass = "c-photo-slider__nav--right";

const slideNavButtonLeftClass = "c-photo-slider__nav-button--left";
const slideNavButtonRightClass = "c-photo-slider__nav-button--right";

const slideShareToolsClass = "c-photo-slider__share-tools";

// component article media image
// const slideMediaClass = "c-article-media__image";
const slideMediaLinkClass = "c-article-media__link";
// const slideMediaLinkLoadedClass = "c-article-media__link--loaded";
const slideMediaInfoClass = "c-article-media__info";
const slideMediaCenterClass = "c-article-media__center";

// component image
const slideImageClass = "c-image";

let McHammerTime1;
let McHammerTime2;

export default {

  init: function(rootElement, rootClass, slideStart = 0) {
    if (!rootElement) {
      return;
    }

    // determine vendor prefix for CSS transform
    this.styleTransform = this.vendorPrefix("Transform");

    this.rootElement = rootElement;
    this.rootClass = rootClass; // js hook class name
    this.slideActive = slideStart; // current slide shown
    this.sliding = false; // block overlaping slide animations

    this.slideSpeed = 200; // ms: speed when animating between slides
    this.slideMargin = 20; // px: margin between slides

    this.landscapeMax = 1020; // px: only applied to landscape photos
    this.portraitMax = 940; // px: only applied to portrait photos
    this.baseScale = 100; // px: default width of photos, then scaled up or down

    this.lockAxis = 0; // enum: x-axis = 1, y-axis = -1, neither = 0
    this.velocityThreshold = 0.5; // swipe velocity threshold for axis - for flick swipes
    this.distanceThreshold = 200; // px: swipe distance threshold for axis - for slow swipes
    this.yAxisMultiplier = 2; // make it require more effort to do y-axis flick/pan swipes

    // unique element id for this slider (that will be prefixed with various string)
    this.uniqueId = (0 | Math.random() * 9e6).toString(36); // e.g. "1p7ym"

    this.slideList = rootElement.getElementsByClassName(slideListClass).item(0);
    this.slideShareTools = rootElement.getElementsByClassName(slideShareToolsClass).item(0);
    this.slideItems = this.arrayFrom(rootElement.getElementsByClassName(slideItemClass));
    this.slides = [];

    if (!this.slideList || !this.slideItems) {
      return;
    }

    // Create array of "slide" objects
    this.slideItems.forEach(item => {
      const slideType = this.arrayFrom(item.classList).reduce((accumulator, className) => {
        if (className.indexOf("c-photo-slider__item--") !== -1) {
          return className.split("--")[1];
        }
      });
      this.slides.push({
        root: item,
        type: slideType,
        image: item.getElementsByClassName(slideImageClass).item(0),
        imageLink: item.getElementsByClassName(slideMediaLinkClass).item(0),
        info: item.getElementsByClassName(slideMediaInfoClass).item(0),
        media: item.getElementsByClassName(slideMediaCenterClass).item(0),
        // grab all images, we don't want any ghosting for any of them
        ghosts: this.arrayFrom(item.getElementsByTagName("img"))
      });
    });

    // target of the left/right button aria-controls
    this.slideList.id = `c-photo-slider-${this.uniqueId}`;

    rootElement.addEventListener("lazybeforesizes", this.onLazyBeforeSizes.bind(this));

    this.setupZoom(); // reset all pinch zoom related vars

    // this will cause placeholder image to be positioned
    if (this.slides[this.slideActive].type === "photo") {
      const placeholderScale = this.scaleImage(this.slideActive, true, false, null);
      this.scaleBase = this.scaleZoom = this.scaleMin = placeholderScale;
    }

    // trigger this manually to get intial measurements
    this.onSliderWindowResize({
      width: document.body.clientWidth
    });

    // go to default slide
    this.slideTo(this.slideActive, false, this.slideWidth);

    window.requestAnimationFrame(this.onHideGhostImages.bind(this));

    // setup aria attributes for links
    this.slides.forEach((slide, index) => {
      if (slide.imageLink) {
        const id = this.slideList.id + "-" + (index + 1);
        slide.imageLink.setAttribute("role", "button");
        slide.imageLink.setAttribute("aria-label", "Toggle section");

        // prevent default anchor click, so that hash is not added to URL
        // this would mess up the history popstate when slider is in dialog
        slide.imageLink.addEventListener("click", e => {
          e.preventDefault();
        });

        // for related information overlay
        if (slide.info) {
          slide.imageLink.setAttribute("aria-expanded", "true");
          slide.imageLink.setAttribute("aria-controls", id);
          slide.info.id = id;
        }
      }
    });

    // previous slide button
    this.slideNavLeft = rootElement.getElementsByClassName(slideNavLeftClass).item(0);
    this.slideNavButtonLeft = rootElement.getElementsByClassName(slideNavButtonLeftClass).item(0);
    if (this.slideNavLeft && this.slideNavButtonLeft) {
      this.slideNavButtonLeft.setAttribute("aria-controls", this.slideList.id);
      this.boundOnSlideNavLeft = this.onSlideNavLeft.bind(this);
      this.slideNavLeft.addEventListener("keydown", this.boundOnSlideNavLeft);
      this.slideNavLeft.addEventListener("click", this.boundOnSlideNavLeft);
    }

    // next slide button
    this.slideNavRight = rootElement.getElementsByClassName(slideNavRightClass).item(0);
    this.slideNavButtonRight = rootElement.getElementsByClassName(slideNavButtonRightClass).item(0);
    if (this.slideNavRight && this.slideNavButtonRight) {
      this.slideNavButtonRight.setAttribute("aria-controls", this.slideList.id);
      this.boundOnSlideNavRight = this.onSlideNavRight.bind(this);
      this.slideNavRight.addEventListener("keydown", this.boundOnSlideNavRight);
      this.slideNavRight.addEventListener("click", this.boundOnSlideNavRight);
    }

    // setup initial left/right button states
    window.requestAnimationFrame(this.updateButtonStates.bind(this));

    // keep slider container width updated
    ee.addListener("windowResize.throttled", this.onSliderWindowResize.bind(this));

    // setup hammer time (aka. party time) ᕕ(⌐■_■)ᕗ ♪
    // attached to root element, to prevent touch blocking by other ui elements.
    McHammerTime1 = new Hammer.Manager(this.rootElement);

    McHammerTime1.add(new Hammer.Pan({
      threshold: 5, // minimum distance required
      direction: Hammer.DIRECTION_ALL
    }));
    McHammerTime1.add(new Hammer.Pinch());

    McHammerTime1.on("panstart", this.onSliderPanStart.bind(this));
    McHammerTime1.on("panmove", this.onSliderPanMove.bind(this));
    McHammerTime1.on("panend", this.onSliderPanEnd.bind(this));

    McHammerTime1.on("pinchstart", this.onSliderPinchStart.bind(this));
    McHammerTime1.on("pinchmove", this.onSliderPinchMove.bind(this));
    McHammerTime1.on("pinchend", this.onSliderPinchEnd.bind(this));

    // seperate hammer time to handle taps only, attached to list element.
    // to prevent unintended tap events bubbling from other ui elements.
    McHammerTime2 = new Hammer.Manager(this.slideList);

    McHammerTime2.add(new Hammer.Tap({
      event: "doubletap", // default values
      interval: 300, // time deplay between taps
      threshold: 20, // minimum distance between points
      posThreshold: 40, // minimum distance between points
      taps: 2
    }));
    McHammerTime2.add(new Hammer.Tap({
      event: "singletap"
    }));

    // Keep track of whether or not the listeners are
    // active to prevent code from running unnecessarily
    this.tapEventsEnabled = true;

    McHammerTime2.get("doubletap").recognizeWith("singletap");
    McHammerTime2.get("singletap").requireFailure("doubletap");

    McHammerTime2.on("singletap", this.onSliderSingleTap.bind(this));
    McHammerTime2.on("doubletap", this.onSliderDoubleTap.bind(this));

    // keypress events for navigation
    this.rootElement.addEventListener("keydown", this.onRootElementKeyDown.bind(this));
  },

  enableTapEvents: function() {
    if (!this.tapEventsEnabled) {
      McHammerTime2.get("singletap").set({ enable: true });
      McHammerTime2.get("doubletap").set({ enable: true });
      this.tapEventsEnabled = true;
    }
  },

  disableTapEvents: function() {
    if (this.tapEventsEnabled) {
      McHammerTime2.get("singletap").set({ enable: false });
      McHammerTime2.get("doubletap").set({ enable: false });
      this.tapEventsEnabled = false;
    }
  },

  // attempt to scale slide image, if within safe index range
  scaleSlide: function(index) {
    const safeIndex = this.safeIndex(this.slides, index);
    if (safeIndex === index) {
      this.scaleImage(safeIndex, true, false, null);
    }
  },

  // intitialize all pinch zoom related vars to defaults
  setupZoom: function() {
    this.zoomed = false; // is the current photo zoomed or not?

    this.scaleBase = 0; // base zoom scale when loaded, and after zoom (completed)
    this.scaleZoom = 0; // current zoomed scale during pinch zoom (uncompleted)

    this.scaleMin = 1; // dynamic minimum zoom scale value (set to initial unzoomed scale of image)
    this.scaleMax = 100; // constant maximum zoom scale value (because reasons)

    this.offsetX = 0; // x-axis offset of image position - current
    this.offsetY = 0; // y-axis offset of image position - current

    this.oldsetX = 0; // x-axis offset of image position - previous
    this.oldsetY = 0; // y-axis offset of image position - previous

    this.centerX = 0; // x-axis position of initial pinch center point
    this.centerY = 0; // y-axis position of initial pinch center point

    this.minsetX = 0; // x-axis minimum permitted offset value
    this.minsetY = 0; // y-axis maximum permitted offset value

    this.maxsetX = 0; // x-axis minimum permitted offset value
    this.maxsetY = 0; // y-axis maximum permitted offset value
  },

  // reset vars and set current image scale
  resetZoom: function() {
    this.setupZoom();
    const imageScale = this.scaleImage(this.slideActive, false, true, null);
    this.scaleBase = this.scaleZoom = this.scaleMin = imageScale;
    this.updateLazyElement(this.slideActive, imageScale);
  },

  onLazyBeforeSizes: function(e) {
    if (e.defaultPrevented) {
      return;
    }
    // find the slide element and object that are associated with this image
    const slideElement = e.target.closest("." + slideItemClass);
    if (!slideElement) {
      return;
    }
    let slideIndex = null;
    this.slides.forEach((slide, index) => {
      if (slide.root === slideElement) {
        slideIndex = index;
      }
    });
    if (slideIndex === this.slideActive && !this.zoomed && !!e.detail.dataAttr) {
      this.scaleImage(slideIndex, false, true, e);
    }
  },

  scaleImage: function(index, imageWidth = false, mediaShow = false, lazyEvent = null) {
    const slide = this.slides[index];
    if (slide.type !== "photo") {
      return 1;
    }
    const container = slide.imageLink;
    const image = slide.image;
    const aspectRatio = parseFloat(image.dataset.aspectRatio || 100); //  portrait [_] < 100 > [___] landscape
    const newWidth = this.calculateSize(container, aspectRatio);
    const newScale = this.calculateScale(newWidth);
    const media = slide.media;

    // must set this to bubble up into lazysizes event capture,
    // so that the image "sizes" attribute value is correctly set,
    // and not just a scaled up version of the image loaded based on the viewport width alone.
    if (lazyEvent) {
      lazyEvent.detail.width = Math.round(newWidth);
    }
    window.requestAnimationFrame(() => {
      if (imageWidth) {
        image.style.width = `${this.baseScale}px`; // retain this value
      }
      image.style[this.styleTransform] = `scale(${newScale})`;

      // update figure caption container to match image width
      if (media && mediaShow) {
        media.style.width = `${newWidth}px`;
        media.classList.remove("u-invisible"); // caption is hidden until image is loaded
      }
    });
    return newScale;
  },

  calculateScale: function(width) {
    return !width ? 1 : (width / this.baseScale);
  },

  calculateWidth: function(scale) {
    return !scale ? this.baseScale : (scale * this.baseScale);
  },

  // update image style width to match detail width, which also matches sizes attribute
  calculateSize: function(container, aspectRatio = 100) {
    if (!container) {
      return 0;
    }
    let maxHeight = container.clientHeight;
    let maxWidth = container.clientWidth;

    // maximum values depending on image orientation
    if (aspectRatio > 100) {
      maxWidth = Math.min(maxWidth, this.landscapeMax); // for landscape images
    } else {
      maxHeight = Math.min(maxHeight, this.portraitMax); // for portrait images
    }
    let ret = maxWidth; // return new width value
    const containerRatio = (maxWidth / maxHeight) * 100; // same formula as back-end

    // basically if display is wider than image, and
    // the image height is no longer contained within display height
    if (containerRatio > aspectRatio) {
      ret = (aspectRatio / containerRatio) * maxWidth;
    }
    return ret;
  },

  // toggle the image figure caption section
  toggleInterface: function() {
    window.requestAnimationFrame(function onSliderToggleInterfaceAnimation() {
      if (this.hasInfoShown()) {
        this.hideInfo();
      } else {
        this.showInfo();
      }
    }.bind(this));
  },

  hasInfoShown: function() {
    return this.rootElement.classList.contains(slideHasInfoClass);
  },

  hasNavShown: function() {
    return this.rootElement.classList.contains(slideHasNavClass);
  },

  showNav: function() {
    // only show navigation if more than one slide present
    if (this.slides.length) {
      this.rootElement.classList.add(slideHasNavClass);
    }
  },

  hideNav: function() {
    this.rootElement.classList.remove(slideHasNavClass);
  },

  showInfo: function() {
    this.rootElement.classList.add(slideHasInfoClass);
    this.slides.map(slide => {
      if (slide.imageLink) {
        slide.imageLink.setAttribute("aria-expanded", "true");
      }
      if (slide.info) {
        slide.info.removeAttribute("aria-hidden");
      }
    });
    ee.emitEvent(`${eventNamespace}:toggleinfo`, [this]);
  },

  hideInfo: function() {
    this.rootElement.classList.remove(slideHasInfoClass);
    this.slides.map(slide => {
      if (slide.imageLink) {
        slide.imageLink.setAttribute("aria-expanded", "false");
      }
      if (slide.info) {
        slide.info.setAttribute("aria-hidden", "true");
      }
    });
    ee.emitEvent(`${eventNamespace}:toggleinfo`, [this]);
  },

  hideShareTools: function() {
    this.rootElement.classList.remove(slideHasShareToolsClass);
    this.slideShareTools.classList.add("u-invisible");
  },

  updateButtonStates: function() {
    if (this.slideActive === 0) {
      this.slideNavButtonLeft.setAttribute("aria-disabled", "true");
      this.slideNavButtonLeft.setAttribute("disabled", "disabled");
    } else {
      this.slideNavButtonLeft.removeAttribute("aria-disabled");
      this.slideNavButtonLeft.removeAttribute("disabled");
    }
    if (this.slideActive === this.slides.length - 1) {
      this.slideNavButtonRight.setAttribute("aria-disabled", "true");
      this.slideNavButtonRight.setAttribute("disabled", "disabled");
    } else {
      this.slideNavButtonRight.removeAttribute("aria-disabled");
      this.slideNavButtonRight.removeAttribute("disabled");
    }
  },

  isKeySpaceOrEnter: function(e) {
    switch (e.keyCode) {
      case 32: // spacebar
      case 13: // enter
        if (!e.metaKey) {
          return true;
        }
        break;
    }
    return false;
  },

  // public method
  // recalc widths and setup slides with new widths, then reset the zoom
  onSliderWindowResize: function(dimensions) {
    // TODO: Figure out why this function is being called with dimensions undefined
    if (!dimensions) {
      return;
    }
    // transition is slowest for touch input
    this.slideWidth = dimensions.width + this.slideMargin;
    // transition is faster for mouse input
    this.clickWidth = this.slideWidth / 10;
    // transition is fastest (instant) for keyboard input
    this.keyWidth = 0;
    // default to slide width when setting up other slide positions
    this.setupSlides(this.slideActive, this.slideWidth);
    this.resetZoom();
  },

  onSliderPanStart: function(e) {
    if (!this.zoomed) {
      this.onSliderPanStartRegular(e);
    }
  },

  onSliderPanMove: function(e) {
    if (this.zoomed) {
      this.onSliderPanMoveZoomed(e);
    } else {
      this.onSliderPanMoveRegular(e);
    }
  },

  onSliderPanEnd: function(e) {
    if (this.zoomed) {
      this.onSliderPanEndZoomed(e);
    } else {
      this.onSliderPanEndRegular(e);
    }
  },

  onSliderPanMoveZoomed: function(e) {
    if (e.direction === Hammer.DIRECTION_NONE) {
      return;
    }
    // continue pan action from previous offset values
    const offsetX = this.oldsetX + (e.deltaX / this.scaleZoom);
    const offsetY = this.oldsetY + (e.deltaY / this.scaleZoom);
    this.positionImage(this.scaleZoom, offsetX, offsetY, true);
  },

  onSliderPanEndZoomed: function() {
    this.positionImage(this.scaleZoom, this.offsetX, this.offsetY, true);
    this.oldsetX = this.offsetX;
    this.oldsetY = this.offsetY;
  },

  positionImage: function(scale = 1, offsetX = 0, offsetY = 0, restrict = false) {
    let oX = offsetX;
    let oY = offsetY;

    // do not allow user to pan off of image region
    // aka: do not allow viewport to leave image region
    if (restrict) {
      if (offsetX > this.minsetX && offsetX < this.maxsetX) {
        // nothing
      } else if (offsetX <= this.minsetX) {
        oX = this.minsetX;
      } else if (offsetX >= this.maxsetX) {
        oX = this.maxsetX;
      }
      if (offsetY > this.minsetY && offsetY < this.maxsetY) {
        // nothing
      } else if (offsetY <= this.minsetY) {
        oY = this.minsetY;
      } else if (offsetY >= this.maxsetY) {
        oY = this.maxsetY;
      }
    }

    this.scaleZoom = scale;
    this.offsetX = oX;
    this.offsetY = oY;

    window.requestAnimationFrame(() => {
      this.imageZoom.style[this.styleTransform] = `scale(${this.scaleZoom}) translate(${this.offsetX}px, ${this.offsetY}px)`;
    });
  },

  onSliderPanStartRegular: function() {
    this.setupSlides(this.slideActive, this.slideWidth);
  },

  onSliderPanMoveRegular: function(e) {
    const absoluteX = Math.abs(e.deltaX);
    const absoluteY = Math.abs(e.deltaY);
    const distX = (absoluteX > (this.distanceThreshold / 8)) ? e.deltaX : 0;
    const distY = (absoluteY > (this.distanceThreshold / 2)) ? e.deltaY : 0;

    this.lockAxis = 0; // no movement
    if (e.direction === Hammer.DIRECTION_LEFT || e.direction === Hammer.DIRECTION_RIGHT) {
      this.lockAxis = 1; // x-axis (user intends to change slide)
    } else if (e.direction === Hammer.DIRECTION_UP || e.direction === Hammer.DIRECTION_DOWN) {
      this.lockAxis = -1; // y-axis (user intends to dismiss/close slider)
    }

    if (this.lockAxis !== 0) {
      window.requestAnimationFrame(function onSliderTouchMoveAnimation() {
        this.translateSlide(this.slideActive, this.lockAxis, distX, distY, this.slideWidth);
      }.bind(this));
    }
  },

  onSliderPanEndRegular: function(e) {
    const absoluteX = Math.abs(e.deltaX);
    const absoluteY = Math.abs(e.deltaY);
    const speedX = Math.abs(e.velocityX);
    const speedY = Math.abs(e.velocityY);

    if (absoluteX > absoluteY) {
      if (e.deltaX > 0) {
        if (this.slideActive > 0 &&
          (absoluteX > this.distanceThreshold || speedX > this.velocityThreshold)) {
          this.slideToTouchEvent(this.slideActive - 1); // right
        } else {
          this.slideToTouchEvent(this.slideActive); // cancel
        }
      } else if (e.deltaX < 0) {
        if (this.slideActive < this.slides.length - 1 &&
          (absoluteX > this.distanceThreshold || speedX > this.velocityThreshold)) {
          this.slideToTouchEvent(this.slideActive + 1); // left
        } else {
          this.slideToTouchEvent(this.slideActive); // cancel
        }
      }
    } else {
      if (e.deltaY < 0) {
        if (absoluteY > (this.distanceThreshold * this.yAxisMultiplier) || speedY > this.velocityThreshold * this.yAxisMultiplier) {
          this.closeAnimation(this.slideActive, e.deltaX, true); // up
        } else {
          this.slideToTouchEvent(this.slideActive); // cancel
        }
      } else if (e.deltaY > 0) {
        if (absoluteY > (this.distanceThreshold * this.yAxisMultiplier) || speedY > this.velocityThreshold * this.yAxisMultiplier) {
          this.closeAnimation(this.slideActive, e.deltaX, false); // down
        } else {
          this.slideToTouchEvent(this.slideActive); // cancel
        }
      }
    }
  },

  // toggle the visibility of the caption section
  onSliderSingleTap: function() {
    this.toggleInterface();
  },

  // reset the slide photo to default position
  onSliderDoubleTap: function() {
    this.slideToTouchEvent(this.slideActive);
  },

  onSliderPinchStart: function(e) {
    this.imageZoom = this.slides[this.slideActive].image;

    // determine pinch center point relative to center of viewport
    const halfWidth = this.rootElement.clientWidth / 2;
    const halfHeight = this.rootElement.clientHeight / 2;

    // pinch center x-axis point relative to center of viewport
    if (e.center.x > halfWidth) {
      this.centerX = e.center.x - halfWidth;
    } else {
      this.centerX = -(halfWidth - e.center.x);
    }

    // pinch center y-axis point relative to center of viewport
    if (e.center.y > halfHeight) {
      this.centerY = e.center.y - halfHeight;
    } else {
      this.centerY = -(halfHeight - e.center.y);
    }
  },

  onSliderPinchMove: function(e) {
    // continue from previous position
    let offsetX = this.oldsetX;
    let offsetY = this.oldsetY;
    let newScale = this.scaleBase * e.scale;

    // use a different scaleMin value depending on pinchout or pinchin
    let scaleMinDirection = (e.additionalEvent === "pinchout") ? this.scaleMin : (this.scaleMin * 1.2);

    // prevent scale from being too small or large
    if (newScale < scaleMinDirection) {
      newScale = this.scaleMin;
      // re-center image, since it is no longer zoomed
      this.slideToTouchEvent(this.slideActive);
      return;
    } else if (newScale > this.scaleMax) {
      newScale = this.scaleMax;
    }

    // do not process if scale did not change
    if ((newScale - this.scaleBase) !== 0) {
      // determine where pinch center point would be at for new zoom scale,
      // relative to the center of the viewport/slider
      const newCenterX = (this.centerX * newScale) / this.scaleBase;
      const newCenterY = (this.centerY * newScale) / this.scaleBase;

      // calc offset values so that new pinch center point location
      // is the same as previous location (relative to center of viewport/slider)
      offsetX += (this.centerX - newCenterX) / newScale;
      offsetY += (this.centerY - newCenterY) / newScale;

      // position image without restricting offset values
      this.positionImage(newScale, offsetX, offsetY, false);
    }
  },

  onSliderPinchEnd: function() {
    this.scaleBase = this.scaleZoom;
    this.zoomed = (this.scaleBase > this.scaleMin);

    // setup min and max values for panning when zoomed
    this.maxsetX = Math.abs((this.imageZoom.width / 2) - ((this.rootElement.clientWidth / this.scaleZoom) / 2));
    this.minsetX = -(this.maxsetX);
    this.maxsetY = Math.abs((this.imageZoom.height / 2) - ((this.rootElement.clientHeight / this.scaleZoom) / 2));
    this.minsetY = -(this.maxsetY);

    // don't recale offset, use previous values, but restrict offset values this time
    this.positionImage(this.scaleZoom, this.offsetX, this.offsetY, true);

    // save offset position
    this.oldsetX = this.offsetX;
    this.oldsetY = this.offsetY;

    // attempt to load a lower/higher quality image, based on the zoom scale
    this.updateLazyElement(this.slideActive, this.scaleZoom);
  },

  // trigger lazysizes update element event
  // will update "sizes" attribute with lazyWidth
  // and display appropriate image from data.srcset
  updateLazyElement: function(index, scale) {
    // trigger lazysizes, to update the image width
    const imageZoom = this.slides[index].image;
    if (imageZoom) {
      const lazyWidth = Math.round(this.calculateWidth(scale));
      // false value is for e.detail.dataAttr value (to be ignored in onLazyBeforeSizes)
      window.lazySizes.autoSizer.updateElem(imageZoom, false, lazyWidth);
    }
  },

  // animate the photo slider to appear that it swipes up or down (dismissed)
  closeAnimation: function(slideActive, offsetX, upward) {
    window.requestAnimationFrame(function onCloseAnimationStart() {
      // hide the ui components
      this.hideShareTools();
      this.hideInfo();
      this.hideNav();
      // perform the translation up or down
      const positionY = (upward ? -1 : 1) * (this.rootElement.clientHeight);
      this.translateSlide(slideActive, -1, offsetX, positionY, this.slideWidth);
    }.bind(this));

    // fire event after above tranlsate is complete
    reqTimeout.setTimeout(function closeAnimationComplete() {
      ee.emitEvent(`${eventNamespace}:swipeup`, [this]);
    }.bind(this), this.slideSpeed / 2);
  },

  // triggered by touch screen input
  slideToTouchEvent: function(index) {
    this.slideTo(index, false, this.slideWidth);
  },

  // triggered by mouse input
  slideToClickEvent: function(index) {
    this.slideTo(index, false, this.clickWidth);
  },

  // triggered by keyboard input
  slideToKeyEvent: function(index) {
    this.slideTo(index, true, this.keyWidth);
  },

  // slide index is zero-based
  // for certain trigger events, you would not want to set focus on the new slide
  // for certain trigger events, animation is instant
  slideTo: function(slideIndex, setFocus, slideWidth) {
    if (this.sliding) {
      return;
    }
    this.sliding = true;

    // validate requested slide index
    this.slideActive = this.safeIndex(this.slides, slideIndex);

    // reset global vars and setup for new image
    this.resetZoom();

    // setup the slide animation move
    window.requestAnimationFrame(function onSlideToAnimation() {
      this.translateSlide(this.slideActive, 0, 0, 0, slideWidth);
      this.updateSlideStates();
      this.updateButtonStates();
      this.scaleSlide(this.slideActive + 1);
      this.scaleSlide(this.slideActive - 1);

      // a setTimeout is needed here, since we want to set focus below, after the slide is fully in view.
      // otherwise the focus would cause the viewport to jump ahead of the animation.
      reqTimeout.setTimeout(function onSlideToTimeout() {
        // only set focus when event did not originate from mouse click
        // so set focus for touch input and key input
        if (setFocus) {
          this.setFocus(this.slideActive);
        }
        ee.emitEvent(`${eventNamespace}:selected`, [this]);
        this.setupSlides(this.slideActive, slideWidth);
        this.sliding = false;
      }.bind(this), this.slideSpeed);
    }.bind(this));
  },

  // public method
  // set keyboard focus on slide item link
  setFocus: function(index) {
    const active = index || this.slideActive;
    if (this.slides[active].imageLink) {
      this.slides[active].imageLink.focus();
    } else if (this.slides[active].type === "ad") {
      // Make sure this element has tabindex="-1"
      this.slides[active].root.getElementsByClassName("c-ad__wrapper")[0].focus();
    }
  },

  // setup the initial positions of the neighbouring slides (excluding the current slide)
  setupSlides: function(slideIndex, slideWidth) {
    const nextX = slideWidth;
    const prevX = -(slideWidth);

    this.slides.forEach((slide, index) => {
      if (index > this.slideActive) {
        slide.root.style.transform = `translate(${nextX}px, 0)`;

      } else if (index < this.slideActive) {
        slide.root.style.transform = `translate(${prevX}px, 0)`;
      }
    });
  },

  // incrementally position the previous/current/next slide(s)
  translateSlide: function(slideIndex, lockAxis, offsetX, offsetY, slideWidth) {
    let oX = (lockAxis === -1) ? 0 : offsetX; // y-axis locked?
    let oY = (lockAxis === 1) ? 0 : offsetY; // x-axis locked?

    const prevIndex = slideIndex - 1;
    const nextIndex = slideIndex + 1;

    const nextX = slideWidth + oX;
    const prevX = -(slideWidth) + oX;

    if (this.slides[slideIndex]) {
      this.slides[slideIndex].root.style.transform = `translate(${oX}px, ${oY}px)`;
    }
    if (this.slides[nextIndex]) {
      this.slides[nextIndex].root.classList.remove("u-invisible");
      this.slides[nextIndex].root.style.transform = `translate(${nextX}px, ${oY}px)`;
    }
    if (this.slides[prevIndex]) {
      this.slides[prevIndex].root.classList.remove("u-invisible");
      this.slides[prevIndex].root.style.transform = `translate(${prevX}px, ${oY}px)`;
    }
  },

  updateSlideStates: function() {
    this.slides.forEach((slide, index) => {
      if (index === this.slideActive) {
        slide.root.removeAttribute("aria-hidden");
        slide.root.classList.remove("u-invisible");
      } else {
        slide.root.setAttribute("aria-hidden", "true");
        slide.root.classList.add("u-invisible");
      }
    });
  },

  onSlideNavLeft: function(e) {
    if (this.hasNavShown()) {
      if (e.type === "click" || (e.type === "keydown" && this.isKeySpaceOrEnter(e))) {
        e.preventDefault();
        e.stopPropagation();
        this.slideToClickEvent(this.slideActive - 1); // previous
      }
    }
  },

  onSlideNavRight: function(e) {
    if (this.hasNavShown()) {
      if (e.type === "click" || (e.type === "keydown" && this.isKeySpaceOrEnter(e))) {
        e.preventDefault();
        e.stopPropagation();
        this.slideToClickEvent(this.slideActive + 1); // next
      }
    }
  },

  // prevent all images from showing ghosts when dragged via mouse
  onHideGhostImages: function() {
    this.slides.forEach(slide => {
      slide.ghosts.forEach(ghost => {
        ghost.setAttribute("draggable", "false"); // prevent ghost image drag via mouse
      });
      if (slide.imageLink) {
        slide.imageLink.setAttribute("draggable", "false"); // prevent ghost link drag via mouse
      }
    });
  },

  onRootElementKeyDown: function(e) {
    let match = false;
    switch (e.keyCode) {
      case 34: // page down
      case 39: // right
      case 40: // down
        match = true;
        this.slideToKeyEvent(this.slideActive + 1); // next
        break;

      case 33: // page up
      case 37: // left
      case 38: // up
        match = true;
        this.slideToKeyEvent(this.slideActive - 1); // previous
        break;

      case 36: // home
        match = true;
        this.slideToKeyEvent(0); // first
        break;

      case 35: // end
        match = true;
        this.slideToKeyEvent(this.slides.length - 1); // last
        break;

      // delegate the slide element events to root element
      case 32: // spacebar
      case 13: // enter
        // allow event to pass through if the focused element
        // is outside of the slide list (i.e. share tools)
        if (document.activeElement.closest(`.${slideListClass}`)) {
          match = true;
          this.toggleInterface();
        }
        break;
    }
    // allow all other key events bubble
    if (match && !e.metaKey) {
      e.preventDefault();
      e.stopPropagation();
    }
  },

  // public method
  // generate HTML for a photo slider component
  generatePhotoSlider: function(galleryPhotos, params) {
    if (!galleryPhotos) {
      return;
    }

    // default options
    const defaults = {
      jsHookClass: "js-c-photo-slider",
      svgSpritePath: "sprite.svg",
      shareToolsClone: true,
      shareToolsClass: "c-share-tools",
      shareToolsItemClass: "c-share-tool",
      shareToolsFilterClass: "js-c-photo-slider-remove"
    };
    const options = this.merge(defaults, params);

    let slideList = ""; // all slide list items
    let adsTotal = 0;
    let adsEncountered = 0; // ad slides shouldn't increment the caption counter
    let index = 0; // index of current slide item

    galleryPhotos.forEach(item => {
      if (item.slideType === "ad") {
        adsTotal++;
      }
    });

    for (let i = 0; i < galleryPhotos.length; i++) {
      if (galleryPhotos[i].slideType && galleryPhotos[i].slideType === "ad") {
        adsEncountered++;
        slideList += `
          <li class="c-photo-slider__item c-photo-slider__item--ad">
            <div class="c-photo-slider__item-content">
              ${galleryPhotos[i].slideMarkup}
            </div>
          </li>`;
      } else {
        let slideImage = this.generateImage(index, galleryPhotos[i].placeholderSrc, galleryPhotos[i].src, galleryPhotos[i].srcset,
          galleryPhotos[i].aspectRatio, galleryPhotos[i].alt);
        let slideCaption = this.generateCaption(index - adsEncountered, galleryPhotos.length - adsTotal, galleryPhotos[i].caption, galleryPhotos[i].credit);
        slideList += this.generateListItem(slideImage, slideCaption);
      }
      index++;
    }

    // get first share tools from [article] page
    // if none found, then photo slider will not have share tools
    let toolsClone = null;
    if (options.shareToolsClone) {
      let shareToolsSource = document.getElementsByClassName(options.shareToolsClass).item(0);
      if (shareToolsSource) {
        toolsClone = this.generateShareTools(shareToolsSource, options.shareToolsItemClass, options.shareToolsFilterClass);
      }
    }

    return this.generateSlideDeck(options.jsHookClass, options.svgSpritePath, slideList, toolsClone);
  },

  // generate HTML for a single slide image
  generateImage: function(index, placeholderSrc, dataSrc, dataSrcSet, aspectRatio, altText) {
    // base64 encode of 1x1px transparent GIF
    const hasSrc = placeholderSrc ? `src="${placeholderSrc}"` : "src=\"data:image/gif;base64,R0lGODlhAQABAIAAAAAAAP///yH5BAEAAAAALAAAAAABAAEAAAIBRAA7\"";
    const hasDataSrc = dataSrc ? `data-src="${dataSrc}"` : "";
    const hasDataSrcSet = dataSrcSet ? `data-srcset="${dataSrcSet}"` : "";
    const hasAspectRatio = `data-aspect-ratio="${aspectRatio}"`;
    const hasAlt = altText ? `alt="${altText}"` : "";
    return `
      <a href="#" class="c-article-media__link">
        <img data-index="${index}"
          class="${slideImageClass} ${lazyConfig.lazyClass}"
          ${hasSrc}
          ${hasDataSrc}
          data-sizes="auto"
          ${hasDataSrcSet}
          ${hasAspectRatio}
          ${hasAlt} />
        <div class="c-spinner c-spinner-lg c-spinner-light"></div>
      </a>`;
  },

  // generate HTML for a single image figure caption
  generateCaption: function(index, total, caption = "", credit = "") {
    const hasMediaCaption = caption ? `<p class="c-article-media__caption">${caption}</p>` : "";
    const hasMediaCredit = credit ? `<p class="c-article-media__credit">${credit}</p>` : "";
    return `
      <figcaption class="c-article-media__info c-article-media__info--dark c-article-media__info--center">
        <div class="c-article-media__center u-invisible">
          <p class="c-article-media__label">${index + 1} of ${total}</p>
          <div class="c-article-media__description">
            ${hasMediaCaption}
            ${hasMediaCredit}
          </div>
        </div>
      </figcaption>`;
  },

  // generate the HTML for a single slide list item
  generateListItem: function(slideImage = "", slideCaption = "") {
    return `
      <li class="c-photo-slider__item c-photo-slider__item--photo">
        <div class="c-photo-slider__item-content">
          <figure class="c-article-media c-article-media--overlay">
            <div class="c-article-media__image c-article-media__image--full-fit">
              ${slideImage}
            </div>
            ${slideCaption}
          </figure>
        </div>
      </li>`;
  },

  // generate the HTML for the share tools section
  // extract the share tools from the article page
  // optionally remove items that contain filterClass
  generateShareTools: function(shareToolsElement, toolClass, filterClass) {
    if (!shareToolsElement) {
      return;
    }

    let shareItems = this.arrayFrom(shareToolsElement.getElementsByClassName(toolClass));
    if (filterClass) {
      shareItems = shareItems.filter((item) => {
        return !item.classList.contains(filterClass);
      });
    }
    shareItems = shareItems.map((item) => {
      return item.outerHTML;
    });

    return `
      <div class="c-photo-slider__share-tools">
        <div class="c-share-tools c-share-tools--dark u-roofless u-floorless">
          ${shareItems.join("")}
        </div>
      </div>`;
  },

  // generate slide deck with: prev/next buttons, slides, and share tools
  generateSlideDeck: function(jsHookClass, svgSpritePath, slidesHtml, shareTools) {
    const hookClass = jsHookClass || "js-c-photo-slider";
    const hasShareTools = shareTools ? "c-photo-slider--has-share-tools" : "";
    return `
      <div class="c-photo-slider ${hookClass} ${hasShareTools} c-photo-slider--has-info c-photo-slider--has-nav"
      role="region" aria-label="Photo slideshow" tabindex="0">
        ${this.generateSlideButton(false, svgSpritePath)}
        ${this.generateSlideButton(true, svgSpritePath)}
        <ul class="c-photo-slider__list" aria-live="polite" aria-atomic="true">
          ${slidesHtml}
        </ul>
        ${shareTools}
      </div>`;
  },

  // generate a prev/next button
  generateSlideButton: function(next, svgSpritePath) {
    const direction = next ? "right" : "left";
    const label = next ? "next" : "previous";
    return `
      <div class="c-photo-slider__nav c-photo-slider__nav--${direction}">
        <button class="c-photo-slider__nav-button c-photo-slider__nav-button--${direction}" aria-label="View ${label} photo">
          <svg class="c-photo-slider__nav-icon" role="img" pointer-events="none" focusable="false" aria-hidden="true">
            <use xlink:href="${svgSpritePath}#icon-${direction}-chevron"></use>
          </svg>
        </button>
      </div>`;
  },

  // simple shallow extend/assign method
  merge: function(obj, src) {
    for (let key in src) {
      if (src.hasOwnProperty(key)) {
        obj[key] = src[key];
      }
    }
    return obj;
  },

  // public method
  // return unique identifier class name
  getRootClass: function() {
    return this.rootClass;
  },

  // public method
  // return current zero-based slide index
  getIndex: function() {
    return this.slideActive;
  },

  getNextIndex: function() {
    let nextIndex = this.slideActive + 1;
    if (!this.slides[nextIndex]) {
      nextIndex = 0;
    }
    return nextIndex;
  },

  getPrevIndex: function() {
    let prevIndex = this.slideActive - 1;
    if (prevIndex < 0) {
      prevIndex = this.slides.length - 1;
    }
    return prevIndex;
  },

  // public method
  // return current slide type
  getSlideType: function() {
    return this.slides[this.slideActive].type;
  },

  getNextSlideType: function() {
    return this.slides[this.getNextIndex()].type;
  },

  getPrevSlideType: function() {
    return this.slides[this.getPrevIndex()].type;
  },

  getNumberOfAdSlidesBeforeCurrent: function() {
    let adCount = 0;
    for (let i = 0; i < this.slideActive + 1; i++) {
      if (this.slides[i].type === "ad") {
        adCount++;
      }
    }
    return adCount;
  },

  // return a safe index value for an array
  // within zero and the array length
  safeIndex: function(list, index) {
    if (!list || !Array.isArray(list) || index === null || typeof index === "undefined") {
      return 0;
    } else if (index < 0) {
      return 0;
    } else if (index >= list.length) {
      return list.length - 1;
    } else {
      return index;
    }
  },

  // simple polyfill for Array.from if neccessary
  arrayFrom: function(HTMLCollection) {
    if (!Array.from) {
      return (HTMLCollection ? [].slice.call(HTMLCollection) : []);
    } else {
      return Array.from(HTMLCollection);
    }
  },

  vendorPrefix: function(property) {
    let testEl = document.createElement("div");
    if (testEl.style.transform === null) {
      let vendors = ["webkit", "moz", "ms", "o"];
      for (let vendor in vendors) {
        if (testEl.style[vendors[vendor] + property] !== undefined) {
          return vendors[vendor] + property;
        }
      }
    }
    return property.toLowerCase();
  }

};
