/**
 * js logic for sticky header
 */

import ee from "~/event-emitter";

const headerNavSelector = ".c-header-nav";
const headerNavFixedClass = "c-header-nav--fixed";
const sectionHeaderSelector = ".c-header-title--section";
const articleHeaderSelector = ".c-header-title--article";
const sectionShowTitleClass = "c-header-title--show";
const sectionUp = "c-header-title--scrolling-up";
const sectionDown = "c-header-title--scrolling-down";
const thresholdFirst = 100;
const thresholdSecond = 300;
const thresholdScrollUp = 50;

let headerNav = null;
let header = null;
let pinned = false;
let classNavBarStatus = 0;
let classSectionStatus = 0;
let distance = null;
let lastScrollTop = 0;
let anchorPoint = null;
let distanceFromAnchor = null;
let panelIsOpen = false;
let scrollPosition = null;
let scrollPercentage = null;

/**
 * @method init
 * @description init method, adds event listener for scroll and resize
 * @return {void}
 */
export default function init() {
  headerNav = document.querySelector(headerNavSelector);
  header = document.querySelector(sectionHeaderSelector) || document.querySelector(articleHeaderSelector);

  // sticky scrolling logic
  let originalScrollValue = window.pageYOffset;

  window.scrollTo(0, 0);
  headerNav && (distance = headerNav.offsetTop - window.pageYOffset);
  window.scrollTo(0, originalScrollValue);

  // run once on initialization
  onViewPortChange();

  ee.on("panel:open", () => {
    panelIsOpen = true;
  });

  ee.on("panel:close", () => {
    panelIsOpen = false;
  });
}

/**
 * @method onViewPortChange
 * @description method checks page scroll offset and applies toggles a class on/off
 * @return {void}
 */
export function onViewPortChange() {
  scrollPosition = window.pageYOffset || document.documentElement.scrollTop;
  scrollPercentage = Math.round(((scrollPosition + window.innerHeight) / document.body.offsetHeight) * 100) / 100;

  if (headerNav) {
    const isFixed = headerNav.classList.contains(headerNavFixedClass);

    if (scrollPosition >= distance && !isFixed && !pinned) {

      pinned = true;
      headerNav.classList.add(headerNavFixedClass);

    } else if (scrollPosition <= distance && pinned) {

      pinned = false;
      headerNav.classList.remove(headerNavFixedClass);

    }
  }

  if (header) {
    const withinTopThreshold = scrollPosition < thresholdFirst;
    const withinSecondThreshold = scrollPosition < thresholdSecond;

    distanceFromAnchor = anchorPoint && Math.abs(scrollPosition - anchorPoint);

    // if within top threshold swap classes to show title or button containers
    if (withinTopThreshold && classSectionStatus !== 0) {

      header.classList.remove(sectionShowTitleClass);
      classSectionStatus = 0;

    } else if (!withinTopThreshold && classSectionStatus !== 1) {

      header.classList.add(sectionShowTitleClass);
      classSectionStatus = 1;

    }

    if (!panelIsOpen) {
      // if within second threshold swap classes to show/hide header-na`
      if (withinSecondThreshold && classNavBarStatus !== 0) {

        resetNav();

      } else if (scrollPercentage === 1 && classNavBarStatus !== 4) {

        scrolledToBottom();

      } else if (!withinSecondThreshold && (scrollPosition > lastScrollTop) && classNavBarStatus !== 1) {

        // scrolling down
        hideNav();

      } else if (!withinSecondThreshold && (scrollPosition < lastScrollTop)) {

        if (classNavBarStatus === 1) {

          // starting to scroll up and set up anchor point
          anchorPoint = scrollPosition;
          classNavBarStatus = 2;
          ee.emit("sticky-header:anchored");

        } else if (classNavBarStatus === 2) {

          // scrolling up and checking distance from anchor
          if (distanceFromAnchor > thresholdScrollUp) {
            showNav();
          }

        }

      }
    }

    lastScrollTop = scrollPosition;
  }
}

/**
 * @method resetNav
 * @description TODO
 * @return {void}
 */
export function resetNav() {
  header.classList.remove(sectionUp);
  header.classList.remove(sectionDown);
  distanceFromAnchor = null;
  classNavBarStatus = 0;
  ee.emit("sticky-header:top");
}

/**
 * @method scrolledToBottom
 * @description TODO
 * @return {void}
 */
export function scrolledToBottom() {
  classNavBarStatus = 4;
  ee.emit("sticky-header:bottom");
}

/**
 * @method showNav
 * @description TODO
 * @return {void}
 */
export function showNav() {
  header.classList.add(sectionUp);
  header.classList.remove(sectionDown);
  classNavBarStatus = 3;
  ee.emit("sticky-header:show");
}

/**
 * @method hideNav
 * @description TODO
 * @return {void}
 */
export function hideNav() {
  header.classList.remove(sectionUp);
  header.classList.add(sectionDown);
  if (classNavBarStatus === 2 || classNavBarStatus === 3) {
    anchorPoint = scrollPosition;
  }
  classNavBarStatus = 1;
  ee.emit("sticky-header:hide");
}
